from ply import lex


class FeatureLexer(object):

    reserved = {
        'and': 'AND',
        'or': 'OR',
        'not': 'NOT',
        "in": 'IN',
        "if": 'IF',
        "then": 'THEN',
        "else": 'ELSE',
    }

    tokens = (
        'HEX',
        'FLOAT',
        'NUMBER',
        'LPAREN',
        'RPAREN',
        'EQUALS',
        'NOT_EQUALS',
        'GT',
        'LT',
        'GTE',
        'LTE',
        'COMMENT',
        'STRING',
        'ID',
        'TRUE',
        'FALSE',
        'DOT',
        'COMMA',
    ) + tuple(reserved.values())

    t_LPAREN = r'\('
    t_RPAREN = r'\)'

    t_EQUALS = r'=='
    t_NOT_EQUALS = r'!='
    t_GT = r'>'
    t_LT = r'<'
    t_GTE = r'>='
    t_LTE = r'<='

    t_ignore = ' \t'

    literals = '+-*/'

    def __init__(self, context=None):
        if context is None:
            context = {}
        self.context = context

    def symbol_lookup(self, id_):
        return self.context.get(id_, id_)

    def t_HEX(self, t):
        r'0x[0-9a-fA-F]+'
        t.value = int(t.value, 16)
        return t

    def t_FLOAT(self, t):
        r'\d+\.\d+'
        t.value = float(t.value)
        return t

    def t_NUMBER(self, t):
        r'\d+'
        t.value = int(t.value)
        return t

    def t_COMMENT(self, t):
        r'\#.*'
        pass
        # No return value. Token discarded

    def t_DOT(self, t):
        r'\.'
        return t

    def t_COMMA(self, t):
        r'\,'
        return t

    def t_TRUE(self, t):
        r'True'
        t.value = True
        return t

    def t_FALSE(self, t):
        r'False'
        t.value = False
        return t

    def t_STRING(self, t):
        r"(?P<quote>['\"]).*?(?P=quote)"
        t.value = t.value[1:-1]
        return t

    def t_ID(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        t.type = self.reserved.get(t.value, 'ID')    # Check for reserved words
        t.value = self.symbol_lookup(t.value)
        return t

    def t_IF(self, t):
        r'if'
        return t

    def t_THEN(self, t):
        r'then'
        return t

    def t_ELSE(self, t):
        r'else'
        return t

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    # Error handling rule
    def t_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    def test(self, data):
        self.lexer.input(data)
        # tok.type, tok.value, tok.lineno, tok.lexpos
        return [tok for tok in self.lexer]



from setuptools import find_packages, setup


VERSION = '0.3.1'

setup(
    name='minilanguage',
    version=VERSION,
    url='https://github.com/fcurella/minilanguage/',
    author='Flavio Curella',
    author_email='flavio.curella@gmail.com',
    description='A minimal DSL for Python',
    license='MIT',
    packages=find_packages(exclude=['*.tests', 'lextab.py', 'parsetab.py']),
    platforms=["any"],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=[
        "ply>=3",
    ],
    test_suite='minilanguage.tests',
)



from ply import yacc


from .lexer import FeatureLexer


class FeatureParser(object):

    precedence = (
        ('left', 'COMMA', 'DOT',),
        ('left', '+', '-'),
        ('left', '*', '/'),
    )

    def __init__(self, context=None):
        if context is None:
            context = {}
        self.context = context
        self.lexer = FeatureLexer(context)

    def p_expression_comment(self, p):
        'expression : expression COMMENT'
        p[0] = p[1]

    def p_expression_plus(self, p):
        '''expression : expression '+' term'''
        p[0] = p[1] + p[3]

    def p_expression_minus(self, p):
        '''expression : expression '-' term'''
        p[0] = p[1] - p[3]

    def p_expression_and(self, p):
        '''expression : expression AND term'''
        p[0] = p[1] and p[3]

    def p_expression_or(self, p):
        '''expression : expression OR term'''
        p[0] = p[1] or p[3]

    def p_expression_equals(self, p):
        '''expression : expression EQUALS term'''
        p[0] = p[1] == p[3]

    def p_expression_not_equals(self, p):
        '''expression : expression NOT_EQUALS term'''
        p[0] = p[1] != p[3]

    def p_expression_gt(self, p):
        '''expression : expression GT term'''
        p[0] = p[1] > p[3]

    def p_expression_lt(self, p):
        '''expression : expression LT term'''
        p[0] = p[1] < p[3]

    def p_expression_gte(self, p):
        '''expression : expression GTE term'''
        p[0] = p[1] >= p[3]

    def p_expression_lte(self, p):
        '''expression : expression LTE term'''
        p[0] = p[1] <= p[3]

    def p_expression_in(self, p):
        '''expression : expression IN term'''
        p[0] = p[1] in p[3]

    def p_expression_if_else(self, p):
        '''expression : expression IF term ELSE term'''
        p[0] = p[1] if p[3] else p[5]

    def p_expression_if_then_else(self, p):
        '''expression : IF expression THEN term ELSE term'''
        p[0] = p[4] if p[2] else p[6]

    def p_expression_times(self, p):
        '''expression : expression '*' term'''
        p[0] = p[1] * p[3]

    def p_expression_div(self, p):
        '''expression : expression '/' term'''
        p[0] = p[1] / p[3]

    def p_expression_term(self, p):
        'expression : term'
        p[0] = p[1]

    def p_tuple(self, p):
        'tuple : term COMMA term'
        p[0] = (p[1], p[3])

    def p_term_tuple(self, p):
        'term : tuple'
        p[0] = p[1]

    def p_term_factor(self, p):
        'term : factor'
        p[0] = p[1]

    def p_term_not(self, p):
        'term : NOT factor'
        p[0] = not p[2]

    def p_factor_method_args(self, p):
        'factor : factor DOT ID LPAREN tuple RPAREN'
        src = p[1]
        attr = p[3]
        arg = p[5]

        p[0] = getattr(src, attr)(*arg)

    def p_factor_method(self, p):
        'factor : factor DOT ID LPAREN factor RPAREN'
        src = p[1]
        attr = p[3]
        arg = p[5]

        p[0] = getattr(src, attr)(arg)

    def p_factor_dot(self, p):
        'factor : factor DOT factor'
        src = p[1]
        attr = p[3]

        try:
            p[0] = src[attr]
        except KeyError:
            p[0] = getattr(src, attr)

    def p_factor_expr(self, p):
        'factor : LPAREN expression RPAREN'
        p[0] = p[2]

    def p_factor_hex(self, p):
        'factor : HEX'
        p[0] = p[1]

    def p_factor_float(self, p):
        'factor : FLOAT'
        p[0] = p[1]

    def p_factor_num(self, p):
        'factor : NUMBER'
        p[0] = p[1]

    def p_factor_id(self, p):
        'factor : ID'
        p[0] = p[1]

    def p_factor_string(self, p):
        'factor : STRING'
        p[0] = p[1]

    def p_factor_true(self, p):
        'factor : TRUE'
        p[0] = p[1]

    def p_factor_false(self, p):
        'factor : FALSE'
        p[0] = p[1]

    # Error rule for syntax errors
    def p_error(self, p):
        print("Syntax error in input!")

    def build(self, **kwargs):
        self.lexer.build(**kwargs)
        self.tokens = self.lexer.tokens
        self.parser = yacc.yacc(module=self, **kwargs)

    def evaluate(self, data, context=None):
        if context is None:
            context = self.context
        self.lexer.context = context
        return self.parser.parse(data)
